"""Cursor for iterating over results."""
# Standard Modules
from itertools import islice
from typing import Callable
from urllib import parse


parse_query_params = lambda url: {
    k: v[0] for k, v in parse.parse_qs(parse.urlparse(url).query).items()
}

parse_paging_data = lambda paging: {
    "previous": parse_query_params(paging["previous"]),
    "next": parse_query_params(paging["next"]),
}

first = lambda xs, i: list(islice(xs, None, i))


class Cursor(object):
    """Cursor for iterating over results."""

    def __init__(
        self,
        request: Callable,
        before,
        after,
        limit,
        batch_size,
    ):

        self.request = request
        self.limit = limit
        self.batch_size = batch_size

        self.i = 0
        self.batch_i = self.i
        self._paging = None
        self.request_batch(before=before, after=after)

    def request_batch(self, before=None, after=None):

        self.batch = self.request(
            before=before,
            after=after,
            limit=min((self.limit - self.i), self.batch_size),
        )

        batch_paging = parse_paging_data(self.batch["paging"])

        if not self._paging:
            self._paging = {"previous": batch_paging["previous"]}

    def __iter__(self):

        return self

    def __next__(self):

        if self.limit and self.i >= self.limit:
            raise StopIteration

        if self.batch is None or self.batch_i >= len(self.batch["data"]):
            next_batch = self._paging["next"]

            if not next_batch:
                raise StopIteration

            self.batch_i = 0

            self.request_batch(
                before=next_batch.get("before"),
                after=next_batch.get("after"),
            )

        x = self.batch["data"][self.batch_i]

        self.i += 1
        self.batch_i += 1

        self._paging["next"] = {"after": x["id"], "limit": self.batch_size}

        return x

    @property
    def paging(self):

        return self._paging



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['cofactr']

package_data = \
{'': ['*']}

install_requires = \
['urllib3>=1.26.9,<2.0.0']

setup_kwargs = {
    'name': 'cofactr',
    'version': '2.0.0',
    'description': 'Client library for accessing Cofactr data.',
    'long_description': '# Cofactr\n\nPython client library for accessing Cofactr.\n\n## Example\n\n```python\nfrom cofactr.graph import GraphAPI\nfrom cofactr.cursor import first\n\ng = GraphAPI()\n\ncursor = g.get_products(\n    query="esp32",\n    fields="id,aliases,labels,statements{spec,assembly},offers",\n    batch_size=10,  # Data is fetched in batches of 10 products.\n    limit=10,  # `list(cursor)` would have at most 10 elements.\n    external=False,\n)\n\ndata = first(cursor, 2)\n\n# To send to web app.\nresponse = {\n    "data": data,\n    # To get the next 10 after the 2 in `data`.\n    "paging": cursor.paging,\n}\n```\n',
    'author': 'Noah Trueblood',
    'author_email': 'noah@cofactr.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/Cofactr/cofactr-client',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.9,<4.0',
}


setup(**setup_kwargs)



"""Cofactr graph API client."""
# Python Modules
from functools import partial
import json
from typing import Literal, Optional

# 3rd Party Modules
import urllib3

from cofactr.cursor import Cursor

Protocol = Literal["http", "https"]


drop_none_values = lambda d: {k: v for k, v in d.items() if v is not None}


def get_products(http, url, query, fields, before, after, limit, external):
    res = http.request(
        "GET",
        f"{url}/products",
        fields=drop_none_values(
            {
                "q": query,
                "fields": fields,
                "before": before,
                "after": after,
                "limit": limit,
                "external": external,
            }
        ),
    )

    data = json.loads(res.data.decode("utf-8"))

    return data


def get_orgs(http, url, query, before, after, limit):
    res = http.request(
        "GET",
        f"{url}/orgs",
        fields=drop_none_values(
            {
                "q": query,
                "before": before,
                "after": after,
                "limit": limit,
            }
        ),
    )

    data = json.loads(res.data.decode("utf-8"))

    return data


class GraphAPI:
    """A client-side representation of the Cofactr graph API."""

    PROTOCOL: Protocol = "https"
    HOST = "graph.cofactr.com"

    def __init__(
        self, protocol: Optional[Protocol] = PROTOCOL, host: Optional[str] = HOST
    ):

        self.url = f"{protocol}://{host}"
        self.http = urllib3.PoolManager()

    def check_health(self):
        """Check the operational status of the service."""

        res = self.http.request("GET", self.url)

        return json.loads(res.data.decode("utf-8"))

    def get_products(
        self,
        query: Optional[str] = None,
        fields: Optional[str] = None,
        before: Optional[str] = None,
        after: Optional[str] = None,
        limit: Optional[int] = None,
        batch_size: int = 10,
        external: Optional[bool] = True,
    ) -> Cursor:
        """Get products.

        Args:
            query: Search query.
            fields: Used to filter properties that the response should contain. A field can be a
                concrete property like "mpn" or an abstract group of properties like "assembly".
                Example: "id,aliases,labels,statements{spec,assembly},offers"
            before: Upper page boundry, expressed as a product ID.
            after: Lower page boundry, expressed as a product ID.
            limit: Restrict the results of the query to a particular number of documents.
            batch_size: The size of each batch of results requested.
            external: Whether to query external sources.
        """

        request = partial(
            get_products,
            http=self.http,
            url=self.url,
            query=query,
            fields=fields,
            external=external,
        )

        return Cursor(
            request=request,
            before=before,
            after=after,
            limit=limit,
            batch_size=batch_size,
        )

    def get_orgs(
        self,
        query: Optional[str] = None,
        before: Optional[str] = None,
        after: Optional[str] = None,
        limit: Optional[int] = None,
        batch_size: int = 10,
    ):
        """Get organizations.

        Args:
            query: Search query.
            before: Upper page boundry, expressed as a product ID.
            after: Lower page boundry, expressed as a product ID.
            limit: Restrict the results of the query to a particular number of documents.
            batch_size: The size of each batch of results requested.
        """

        request = partial(
            get_orgs,
            http=self.http,
            url=self.url,
            query=query,
        )

        return Cursor(
            request=request,
            before=before,
            after=after,
            limit=limit,
            batch_size=batch_size,
        )

    def get_product(
        self,
        id: str,
        fields: Optional[str] = None,
        external: Optional[bool] = True,
    ):
        """Get product.

        Args:
            fields: Used to filter properties that the response should contain. A field can be a
                concrete property like "mpn" or an abstract group of properties like "assembly".
                Example: "id,aliases,labels,statements{spec,assembly},offers"
            external: Whether to query external sources in order to update information for the
                given product.
        """

        res = self.http.request(
            "GET",
            f"{self.url}/products/{id}",
            fields=drop_none_values(
                {
                    "fields": fields,
                    "external": external,
                }
            ),
        )

        return json.loads(res.data.decode("utf-8"))

    def get_org(self, id: str):
        """Get organization."""

        res = self.http.request("GET", f"{self.url}/orgs/{id}")

        return json.loads(res.data.decode("utf-8"))
